// *********************
// * Da Punisha Plugin *
// *********************
// version 1.4
// by Pierre-Marie Baty <pm@bots-united.com>


#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>

#define NOT_PUNISHED 0
#define IS_CHICKEN 1
#define IS_DUMPING_CRAP 2
#define HAS_BOMB (1 << 6)

#define TEAM_CS_TERRORIST 1
#define TEAM_CS_COUNTER 2

typedef struct
{
   int punish_level;
   edict_t *pChicken;
} player;

enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
DLL_FUNCTIONS gFunctionTable;
cvar_t chicken_punish = {"chicken_punish", "1", FCVAR_EXTDLL};
cvar_t auto_punish = {"auto_punish", "0", FCVAR_EXTDLL};
cvar_t bot_punish = {"bot_punish", "0", FCVAR_EXTDLL};
cvar_t dumpcrap_time = {"dumpcrap_time", "0.1", FCVAR_EXTDLL};
player players[33];

void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals);
void DispatchTouch (edict_t *pentTouched, edict_t *pentOther);
void ServerActivate (edict_t *pEdictList, int edictCount, int clientMax);
void ClientDisconnect (edict_t *pEntity);
void ClientPutInServer (edict_t *pEntity);
void ClientCommand (edict_t *pEntity);
void StartFrame (void);
void PlayerPunish (void);
void PlayerForgive (void);
void engClientCommand (edict_t *pEdict, char *szFmt, ...);
void EmitSound (edict_t *entity, int channel, const char *sample, float volume, float attenuation, int fFlags, int pitch);
Vector UTIL_WrapAngles (Vector angles_to_wrap);
int GetTeam (edict_t *pEntity);
void GetGameDir (char *game_dir);


// START of Metamod stuff

enginefuncs_t meta_engfuncs;
gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;
meta_globals_t *gpMetaGlobals;

META_FUNCTIONS gMetaFunctionTable =
{
   NULL, // pfnGetEntityAPI()
   NULL, // pfnGetEntityAPI_Post()
   GetEntityAPI2, // pfnGetEntityAPI2()
   NULL, // pfnGetEntityAPI2_Post()
   NULL, // pfnGetNewDLLFunctions()
   NULL, // pfnGetNewDLLFunctions_Post()
   GetEngineFunctions, // pfnGetEngineFunctions()
   NULL, // pfnGetEngineFunctions_Post()
};

plugin_info_t Plugin_info = {
   META_INTERFACE_VERSION, // interface version
   "Punish", // plugin name
   "1.4", // plugin version
   "23/04/2004", // date of creation
   "Pierre-Marie Baty <pm@bots-united.com>", // plugin author
   "http://racc.bots-united.com", // plugin URL
   "PUNISH", // plugin logtag
   PT_CHANGELEVEL, // when loadable
   PT_ANYTIME, // when unloadable
};


C_DLLEXPORT int Meta_Query (char *ifvers, plugin_info_t **pPlugInfo, mutil_funcs_t *pMetaUtilFuncs)
{
   // this function is the first function ever called by metamod in the plugin DLL. Its purpose
   // is for metamod to retrieve basic information about the plugin, such as its meta-interface
   // version, for ensuring compatibility with the current version of the running metamod.

   // keep track of the pointers to metamod function tables metamod gives us
   gpMetaUtilFuncs = pMetaUtilFuncs;
   *pPlugInfo = &Plugin_info;

   // check for interface version compatibility
   if (strcmp (ifvers, Plugin_info.ifvers) != 0)
   {
      int mmajor = 0, mminor = 0, pmajor = 0, pminor = 0;

      LOG_CONSOLE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);
      LOG_MESSAGE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);

      // if plugin has later interface version, it's incompatible (update metamod)
      sscanf (ifvers, "%d:%d", &mmajor, &mminor);
      sscanf (META_INTERFACE_VERSION, "%d:%d", &pmajor, &pminor);

      if ((pmajor > mmajor) || ((pmajor == mmajor) && (pminor > mminor)))
      {
         LOG_CONSOLE (PLID, "metamod version is too old for this plugin; update metamod");
         LOG_ERROR (PLID, "metamod version is too old for this plugin; update metamod");
         return (FALSE);
      }

      // if plugin has older major interface version, it's incompatible (update plugin)
      else if (pmajor < mmajor)
      {
         LOG_CONSOLE (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         LOG_ERROR (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         return (FALSE);
      }
   }

   return (TRUE); // tell metamod this plugin looks safe
}


C_DLLEXPORT int Meta_Attach (PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs)
{
   // this function is called when metamod attempts to load the plugin. Since it's the place
   // where we can tell if the plugin will be allowed to run or not, we wait until here to make
   // our initialization stuff, like registering CVARs and dedicated server commands.

   char dirname[256];

   // are we allowed to load this plugin now ?
   if (now > Plugin_info.loadable)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // keep track of the pointers to engine function tables metamod gives us
   gpMetaGlobals = pMGlobals;
   memcpy (pFunctionTable, &gMetaFunctionTable, sizeof (META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;

   // get the MOD directory name (full path)
   GetGameDir (dirname);

   // is it NOT the Counter-Strike directory ?
   if (strcmp (dirname, "cstrike") != 0)
   {
      // then don't allow this plugin to run
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (bad dirname: '%s')", Plugin_info.name, dirname);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (bad dirname: '%s')", Plugin_info.name, dirname);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // print a message to notify about plugin attaching
   LOG_CONSOLE (PLID, "%s: plugin attaching", Plugin_info.name);
   LOG_MESSAGE (PLID, "%s: plugin attaching", Plugin_info.name);

   // ask the engine to register the server commands this plugin uses
   REG_SVR_COMMAND ("punish", PlayerPunish);
   REG_SVR_COMMAND ("forgive", PlayerForgive);

   // ask the engine to register the CVARs this plugin uses
   CVAR_REGISTER (&chicken_punish);
   CVAR_REGISTER (&auto_punish);
   CVAR_REGISTER (&bot_punish);
   CVAR_REGISTER (&dumpcrap_time);

   return (TRUE); // returning TRUE enables metamod to attach this plugin
}


C_DLLEXPORT int Meta_Detach (PLUG_LOADTIME now, PL_UNLOAD_REASON reason)
{
   // this function is called when metamod unloads the plugin. A basic check is made in order
   // to prevent unloading the plugin if its processing should not be interrupted.

   int player_index;
   edict_t *pPlayer;

   // is metamod allowed to unload the plugin ?
   if ((now > Plugin_info.unloadable) && (reason != PNL_CMD_FORCED))
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from unloading this plugin
   }

   // cycle through all players and forgive them as plugin is detaching...
   for (player_index = 1; player_index <= gpGlobals->maxClients; player_index++)
   {
      pPlayer = INDEXENT (player_index);

      if (FNullEnt (pPlayer))
         continue; // skip invalid players

      // "is" this player a chicken ?
      if (!FNullEnt (players[player_index].pChicken))
         players[player_index].pChicken->v.flags |= FL_KILLME; // mark the chicken entity for removal

      // is this player still alive ?
      if (pPlayer->v.health > 0)
         pPlayer->v.effects &= ~EF_NODRAW; // draw the player again
   }

   return (TRUE); // returning TRUE enables metamod to unload this plugin
}


C_DLLEXPORT int GetEntityAPI2 (DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
   gFunctionTable.pfnTouch = DispatchTouch;
   gFunctionTable.pfnClientDisconnect = ClientDisconnect;
   gFunctionTable.pfnClientPutInServer = ClientPutInServer;
   gFunctionTable.pfnClientCommand = ClientCommand;
   gFunctionTable.pfnServerActivate = ServerActivate;
   gFunctionTable.pfnStartFrame = StartFrame;

   memcpy (pFunctionTable, &gFunctionTable, sizeof (DLL_FUNCTIONS));
   return (TRUE);
}


C_DLLEXPORT int GetEngineFunctions (enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion)
{
   meta_engfuncs.pfnEmitSound = EmitSound;
   meta_engfuncs.pfnClientCommand = engClientCommand;

   memcpy (pengfuncsFromEngine, &meta_engfuncs, sizeof (enginefuncs_t));
   return TRUE;
}

// END of Metamod stuff




#ifdef _WIN32
BOOL WINAPI DllMain (HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
   return TRUE; // required Win32 DLL entry point
}
#endif


void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   memcpy (&g_engfuncs, pengfuncsFromEngine, sizeof (enginefuncs_t));
   gpGlobals = pGlobals;
}


void DispatchTouch (edict_t *pentTouched, edict_t *pentOther)
{
   int player_index;

   player_index = ENTINDEX (pentOther);

   if (players[player_index].punish_level == IS_CHICKEN)
      RETURN_META (MRES_SUPERCEDE); // don't allow chickens to touch anything

   RETURN_META (MRES_IGNORED); // normal players touch whatever they can
}


void ServerActivate (edict_t *pEdictList, int edictCount, int clientMax)
{
   // precache chicken stuff
   PRECACHE_MODEL ("models/chick.mdl");
   PRECACHE_SOUND ("misc/killChicken.wav");

   RETURN_META (MRES_IGNORED);
}


void ClientDisconnect (edict_t *pEntity)
{
   int player_index;

   player_index = ENTINDEX (pEntity);

   // erase any of this player's chicken
   if (!FNullEnt (players[player_index].pChicken))
      REMOVE_ENTITY (players[player_index].pChicken);

   players[player_index].pChicken = NULL;
   players[player_index].punish_level = NOT_PUNISHED; // set his punish state to 0

   RETURN_META (MRES_IGNORED);
}


void ClientPutInServer (edict_t *pEntity)
{  
   int player_index;

   player_index = ENTINDEX (pEntity);

   // erase any of this new player's chicken
   if (players[player_index].pChicken != NULL)
      REMOVE_ENTITY (players[player_index].pChicken);

   players[player_index].pChicken = NULL;
   players[player_index].punish_level = NOT_PUNISHED; // set his punish state to 0

   RETURN_META (MRES_IGNORED);
}


void ClientCommand (edict_t *pEntity)
{
   int player_index;

   player_index = ENTINDEX (pEntity);

   if ((players[player_index].punish_level == IS_CHICKEN) && !(pEntity->v.weapons & HAS_BOMB))
      RETURN_META (MRES_SUPERCEDE); // don't allow chickens to issue client commands unless for dropping the bomb

   RETURN_META (MRES_IGNORED); // normal players issue whatever command they want
}


void StartFrame (void)
{
   static float crap_time = 0;
   edict_t *pPlayer;
   int player_index;
   int killer_index;

   // is auto-punishing enabled ?
   if (CVAR_GET_FLOAT ("auto_punish") == 1)
   {
      // cycle through all players
      for (player_index = 1; player_index <= gpGlobals->maxClients; player_index++)
      {
         pPlayer = INDEXENT (player_index);

         if (FNullEnt (pPlayer))
            continue; // skip invalid players

         // is this player dead AND was he killed by another entity ?
         if ((pPlayer->v.deadflag != DEAD_NO) && !FNullEnt (pPlayer->v.dmg_inflictor))
         {
            // is the damage inflictor not a player ?
            if (strcmp (STRING (pPlayer->v.dmg_inflictor->v.classname), "player") != 0)
               continue;

            // is the damage inflictor a bot and is bot punishment disabled ?
            if ((pPlayer->v.dmg_inflictor->v.flags & FL_FAKECLIENT) && (CVAR_GET_FLOAT ("bot_punish") == 0))
               continue; // skip bots

            killer_index = ENTINDEX (pPlayer->v.dmg_inflictor);

            // was this a team kill ?
            if (GetTeam (pPlayer->v.dmg_inflictor) == GetTeam (pPlayer))
            {
               // do we want our llama to be turned into a chicken ?
               if (CVAR_GET_FLOAT ("chicken_punish") == 1)
               {
                  players[killer_index].punish_level = IS_CHICKEN; // turn this teamkiller into a chicken
                  ALERT (at_console, "Player %s TEAM KILLED %s and was turned into a chicken for punishment\n", STRING (pPlayer->v.dmg_inflictor->v.netname), STRING (pPlayer->v.netname));

                  players[killer_index].pChicken = CREATE_NAMED_ENTITY (MAKE_STRING ("info_target")); // create chicken entity
                  players[killer_index].pChicken->v.solid = SOLID_NOT; // don't allow it to be solid
                  players[killer_index].pChicken->v.takedamage = DAMAGE_NO; // don't allow it to take damage
                  players[killer_index].pChicken->v.movetype = MOVETYPE_FLY; // fly-type movement
                  players[killer_index].pChicken->v.classname = MAKE_STRING ("func_breakable"); // sets a name for it
                  players[killer_index].pChicken->v.rendermode = kRenderNormal; // normal rendering mode
                  players[killer_index].pChicken->v.renderfx = kRenderFxNone; // no special FX
                  players[killer_index].pChicken->v.owner = pPlayer; // sets the player as the owner of that chicken
                  SET_MODEL (players[killer_index].pChicken, "models/chick.mdl"); // sets it a model

                  break;
               }

               // else let's dump crap on his hard disk
               else
               {
                  players[killer_index].punish_level = IS_DUMPING_CRAP; // dump crap on his disk
                  ALERT (at_console, "Player %s TEAM KILLED %s and is dumping crap on his hard disk for punishment\n", STRING (pPlayer->v.dmg_inflictor->v.netname), STRING (pPlayer->v.netname));

                  break;
               }
            }
         }
      }
   }

   // cycle through all players again
   for (player_index = 1; player_index <= gpGlobals->maxClients; player_index++)
   {
      pPlayer = INDEXENT (player_index);

      if (FNullEnt (pPlayer))
         continue; // skip invalid players

      // is this player punished and turned into a chicken ?
      if ((players[player_index].punish_level == IS_CHICKEN) && !FNullEnt (players[player_index].pChicken))
      {
         if (pPlayer->v.health > 10)
         {
            pPlayer->v.health = 10; // don't allow player to have more than 10 health
            MESSAGE_BEGIN (MSG_ONE, GET_USER_MSG_ID (PLID, "Health", NULL), NULL, pPlayer);
            WRITE_BYTE (pPlayer->v.health); // let other players know this new value
            MESSAGE_END ();
         }

         // if this player has to be turned into a chicken then let him drop the C4 if needed
         if ((players[player_index].punish_level == IS_CHICKEN) && (pPlayer->v.weapons & HAS_BOMB))
         {
            if (crap_time < gpGlobals->time)
            {
               engClientCommand (pPlayer, "weapon_c4\n"); // if this player owns the C4
               engClientCommand (pPlayer, "drop\n"); // then drop it
               crap_time = gpGlobals->time + 0.2; // next command in 0.2 second
            }
            continue; // skip this player since he hasn't dropped the C4 yet
         }

         pPlayer->v.effects |= EF_NODRAW; // make this player invisible

         MESSAGE_BEGIN (MSG_ONE, GET_USER_MSG_ID (PLID, "ResetHUD", NULL), NULL, pPlayer);
         MESSAGE_END ();

         pPlayer->v.viewmodel = 0; // don't show him his weapon
         pPlayer->v.weapons = 0; // clear his weapons
         pPlayer->v.view_ofs = Vector (0, 0, -10); // set the player's view offset to his chicken's origin

         // is this player alive ?
         if (pPlayer->v.deadflag != DEAD_NO)
            players[player_index].pChicken->v.effects |= EF_NODRAW; // draw his chicken
         else
            players[player_index].pChicken->v.effects &= ~EF_NODRAW; // don't draw his chicken

         players[player_index].pChicken->v.health = pPlayer->v.health; // update chicken's health
         players[player_index].pChicken->v.origin = pPlayer->v.origin + Vector (0, 0, -36); // set his chicken's origin to where he is
         players[player_index].pChicken->v.angles = UTIL_WrapAngles (pPlayer->v.angles + Vector (0, -90, 0)); // set his chicken's angles to his body angles
      }

      // else is this player punished and dumping crap on his disk ?
      else if (players[player_index].punish_level == IS_DUMPING_CRAP)
      {
         if (crap_time < gpGlobals->time)
         {
            engClientCommand (pPlayer, "snapshot\n"); // make this player take a snapshot hehe...
            crap_time = gpGlobals->time + CVAR_GET_FLOAT ("dumpcrap_time"); // next snapshot in some time
         }
      }
   }

   RETURN_META (MRES_IGNORED);
}


void PlayerPunish (void)
{
   edict_t *pPlayer;
   int player_index;

   // cycle through all players
   for (player_index = 1; player_index <= gpGlobals->maxClients; player_index++)
   {
      pPlayer = INDEXENT (player_index);

      if (FNullEnt (pPlayer))
         continue; // skip invalid players

      // have we found the player we want to punish ?
      if (strcmp (STRING (pPlayer->v.netname), CMD_ARGV (1)) == 0)
      {
         // do we want our llama to be turned into a chicken ?
         if (CVAR_GET_FLOAT ("chicken_punish") == 1)
         {
            players[player_index].punish_level = IS_CHICKEN; // turn this player into a chicken
            ALERT (at_console, "Player %s has been turned into a chicken for punishment\n", CMD_ARGV (1));

            players[player_index].pChicken = CREATE_NAMED_ENTITY (MAKE_STRING ("info_target")); // create chicken entity
            players[player_index].pChicken->v.solid = SOLID_NOT; // don't allow it to be solid
            players[player_index].pChicken->v.takedamage = DAMAGE_NO; // don't allow it to take damage
            players[player_index].pChicken->v.movetype = MOVETYPE_FLY; // fly-type movement
            players[player_index].pChicken->v.classname = MAKE_STRING ("func_breakable"); // sets a name for it
            players[player_index].pChicken->v.rendermode = kRenderNormal; // normal rendering mode
            players[player_index].pChicken->v.renderfx = kRenderFxNone; // no special FX
            players[player_index].pChicken->v.owner = pPlayer; // sets the player as the owner of that chicken
            SET_MODEL (players[player_index].pChicken, "models/chick.mdl"); // sets it a model
         }

         // else let's dump crap on his hard disk
         else
         {
            players[player_index].punish_level = IS_DUMPING_CRAP; // turn this teamkiller into a chicken
            ALERT (at_console, "Player %s is dumping crap on his hard disk for punishment\n", CMD_ARGV (1));
         }

         return;
      }
   }

   ALERT (at_console, "Alert: Player \"%s\" was not found!\n", CMD_ARGV (1));
   return;
}


void PlayerForgive (void)
{
   edict_t *pPlayer;
   int player_index;

   // cycle through all players
   for (player_index = 1; player_index <= gpGlobals->maxClients; player_index++)
   {
      pPlayer = INDEXENT (player_index);

      if (FNullEnt (pPlayer))
         continue; // skip invalid players

      // have we found the player we want to punish ?
      if (strcmp (STRING (pPlayer->v.netname), CMD_ARGV (1)) == 0)
      {
         ALERT (at_console, "Player %s has been forgiven\n", CMD_ARGV (1));

         // remove any of this new player's chicken
         if (players[player_index].pChicken != NULL)
         {
            REMOVE_ENTITY (players[player_index].pChicken);
            players[player_index].pChicken = NULL;
            pPlayer->v.effects &= ~EF_NODRAW; // draw it again
         }

         players[player_index].punish_level = NOT_PUNISHED; // set his punish state to 0
         MDLL_ClientKill (pPlayer); // kill him

         return;
      }
   }

   ALERT (at_console, "Alert: Player \"%s\" was not found!\n", CMD_ARGV (1));
   return;
}


void engClientCommand (edict_t *pEdict, char *szFmt, ...)
{
   va_list ap;
   char buf[1024];

   va_start (ap, szFmt);
   vsnprintf (buf, sizeof (buf), szFmt, ap);
   va_end (ap);

   (*g_engfuncs.pfnClientCommand) (pEdict, buf);

   RETURN_META (MRES_IGNORED);
}


void EmitSound (edict_t *entity, int channel, const char *sample, float volume, float attenuation, int fFlags, int pitch)
{
   int player_index;

   player_index = ENTINDEX (entity);

   // is this player a chicken ?
   if ((player_index > 0) && (player_index <= gpGlobals->maxClients)
       && (players[player_index].punish_level == IS_CHICKEN))
   {
      (*g_engfuncs.pfnEmitSound) (entity, channel, "misc/killChicken.wav", volume, attenuation, fFlags, RANDOM_FLOAT (90, 110));
      RETURN_META (MRES_SUPERCEDE); // every sound he makes is a chicken sound
   }

   RETURN_META (MRES_IGNORED);
}


Vector UTIL_WrapAngles (Vector angles)
{
   // check for wraparound of angles
   if (angles.x >= 180)
      angles.x -= 360 * ((int) (angles.x / 360) + 1); // and not 0.5
   if (angles.x < -180)
      angles.x += 360 * ((int) (-angles.x / 360) + 1); // and not 0.5
   if (angles.y >= 180)
      angles.y -= 360 * ((int) (angles.y / 360) + 1); // and not 0.5
   if (angles.y < -180)
      angles.y += 360 * ((int) (-angles.y / 360) + 1); // and not 0.5
   angles.z = 0.0;

   if (angles.x > 89)
      angles.x = 89;
   else if (angles.x < -89)
      angles.x = -89;

   if ((angles.x >= 180) || (angles.x < -180))
      angles.x = 0; // heck, if we're still above the limit then something's REALLY fuckedup!
   if ((angles.y >= 180) || (angles.y < -180))
      angles.y = 0; // heck, if we're still above the limit then something's REALLY fuckedup!

   return (angles);
}


int GetTeam (edict_t *pEntity)
{
   char model_name[32];

   strcpy (model_name, (INFOKEY_VALUE (GET_INFOKEYBUFFER (pEntity), "model")));

   if ((model_name[2] == 'b') // urBan
       || (model_name[2] == 'g') // gsG9, giGn
       || (model_name[2] == 's') // saS
       || (model_name[2] == 'p')) // viP
      return (TEAM_CS_COUNTER);

   return (TEAM_CS_TERRORIST); // teRror, leEt, guErilla, arCtic
}


void GetGameDir (char *game_dir)
{
   // This function fixes the erratic behaviour caused by the use of the GET_GAME_DIR engine
   // macro, which returns either an absolute directory path, or a relative one, depending on
   // whether the game server is run standalone or not. This one always return a RELATIVE path.

   unsigned char length, fieldstart, fieldstop;

   GET_GAME_DIR (game_dir); // call the engine macro and let it mallocate for the char pointer

   length = strlen (game_dir); // get the length of the returned string
   length--; // ignore the trailing string terminator

   // format the returned string to get the last directory name
   fieldstop = length;
   while (((game_dir[fieldstop] == '\\') || (game_dir[fieldstop] == '/')) && (fieldstop > 0))
      fieldstop--; // shift back any trailing separator

   fieldstart = fieldstop;
   while ((game_dir[fieldstart] != '\\') && (game_dir[fieldstart] != '/') && (fieldstart > 0))
      fieldstart--; // shift back to the start of the last subdirectory name

   if ((game_dir[fieldstart] == '\\') || (game_dir[fieldstart] == '/'))
      fieldstart++; // if we reached a separator, step over it

   // now copy the formatted string back onto itself character per character
   for (length = fieldstart; length <= fieldstop; length++)
      game_dir[length - fieldstart] = game_dir[length];
   game_dir[length - fieldstart] = 0; // terminate the string

   return;
}
